package java.util.concurrent.atomic;

import java.io.IOException;
import java.io.InvalidObjectException;
import java.lang.reflect.Array;
import java.io.ObjectInputStream;
import java.util.function.BinaryOperator;
import java.util.function.UnaryOperator;
import java.util.Arrays;
import sun.misc.Unsafe;
import java.io.Serializable;

public class AtomicReferenceArray<E> implements Serializable
{
    private static final long serialVersionUID = -6209656149925076980L;
    private static final Unsafe unsafe;
    private static final int base;
    private static final int shift;
    private static final long arrayFieldOffset;
    private final Object[] array;
    
    private long checkedByteOffset(int i) {
        if (i < 0 || i >= this.array.length) {
            throw new IndexOutOfBoundsException("index " + i);
        }
        return byteOffset(i);
    }
    
    private static long byteOffset(int n) {
        return ((long)n << AtomicReferenceArray.shift) + (long)AtomicReferenceArray.base;
    }
    
    public AtomicReferenceArray(int n) {
        this.array = new Object[n];
    }
    
    public AtomicReferenceArray(E[] original) {
        this.array = Arrays.copyOf(original, original.length, Object[].class);
    }
    
    public final int length() {
        return this.array.length;
    }
    
    public final E get(int n) {
        return this.getRaw(this.checkedByteOffset(n));
    }
    
    private E getRaw(long offset) {
        return AtomicReferenceArray.unsafe.getObjectVolatile(this.array, offset);
    }
    
    public final void set(int n, E x) {
        AtomicReferenceArray.unsafe.putObjectVolatile(this.array, this.checkedByteOffset(n), x);
    }
    
    public final void lazySet(int n, E x) {
        AtomicReferenceArray.unsafe.putOrderedObject(this.array, this.checkedByteOffset(n), x);
    }
    
    public final E getAndSet(int n, E newValue) {
        return AtomicReferenceArray.unsafe.getAndSetObject(this.array, this.checkedByteOffset(n), newValue);
    }
    
    public final boolean compareAndSet(int n, E e, E e2) {
        return this.compareAndSetRaw(this.checkedByteOffset(n), e, e2);
    }
    
    private boolean compareAndSetRaw(long offset, E expected, E x) {
        return AtomicReferenceArray.unsafe.compareAndSwapObject(this.array, offset, expected, x);
    }
    
    public final boolean weakCompareAndSet(int n, E e, E e2) {
        return this.compareAndSet(n, e, e2);
    }
    
    public final E getAndUpdate(int n, UnaryOperator<E> unaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        Object raw;
        do {
            raw = this.getRaw(checkedByteOffset);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, unaryOperator.apply(raw)));
        return raw;
    }
    
    public final E updateAndGet(int n, UnaryOperator<E> unaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        Object raw;
        Object apply;
        do {
            raw = this.getRaw(checkedByteOffset);
            apply = unaryOperator.apply(raw);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, apply));
        return apply;
    }
    
    public final E getAndAccumulate(int n, E e, BinaryOperator<E> binaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        Object raw;
        do {
            raw = this.getRaw(checkedByteOffset);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, binaryOperator.apply(raw, e)));
        return raw;
    }
    
    public final E accumulateAndGet(int n, E e, BinaryOperator<E> binaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        Object raw;
        Object apply;
        do {
            raw = this.getRaw(checkedByteOffset);
            apply = binaryOperator.apply(raw, e);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, apply));
        return apply;
    }
    
    @Override
    public String toString() {
        int n = this.array.length - 1;
        if (n == -1) {
            return "[]";
        }
        StringBuilder sb = new StringBuilder();
        sb.append('[');
        int n2 = 0;
        while (true) {
            sb.append(this.getRaw(byteOffset(n2)));
            if (n2 == n) {
                break;
            }
            sb.append(',').append(' ');
            ++n2;
        }
        return sb.append(']').toString();
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException, InvalidObjectException {
        Object x = objectInputStream.readFields().get("array", null);
        if (x == null || !x.getClass().isArray()) {
            throw new InvalidObjectException("Not array type");
        }
        if (x.getClass() != Object[].class) {
            x = Arrays.copyOf((U[])(U[])x, Array.getLength(x), Object[].class);
        }
        AtomicReferenceArray.unsafe.putObjectVolatile(this, AtomicReferenceArray.arrayFieldOffset, x);
        return;
    }
    
    static {
        try {
            unsafe = Unsafe.getUnsafe();
            arrayFieldOffset = AtomicReferenceArray.unsafe.objectFieldOffset(AtomicReferenceArray.class.getDeclaredField("array"));
            base = AtomicReferenceArray.unsafe.arrayBaseOffset(Object[].class);
            int arrayIndexScale = AtomicReferenceArray.unsafe.arrayIndexScale(Object[].class);
            if ((arrayIndexScale & arrayIndexScale - 1) != 0x0) {
                throw new Error("data type scale not a power of two");
            }
            shift = 31 - Integer.numberOfLeadingZeros(arrayIndexScale);
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
}
